#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

/* This is an evil cludge just because something seems to be screwy
   with my include files on my Linux boxen... */
#ifndef popen
FILE *popen(const char *command, const char *type);
int pclose(FILE *stream);
#endif

#define MP3DIR "/misc/mp3/"	/* Where you have your mp3s */
#define MP3MD5 "/misc/mp3md5"	/* Where the program will store its file */
#define FNLEN		1000	/* Max length of a filename (cludge) */
#define MD5LEN		100	/* Max length of a md5-result */
#define SKIP		10000	/* How many bytes to skip in start of file */
#define CHECKLEN	30000	/* How much of the mp3 to be md5:ed
				   -1 means all of it */
#define CHECKMIN	30000	/* Minimum amount of bytes to be checked */

int show_matches = 1;
int show_no_matches = 1;

/* ======================================================================
   trim_file: Trim the file from id3-tags and stuff 
   ====================================================================== */

off_t trim_file (int mp3, char *fn)
{
  char buf[3];
  off_t total;
  
  /* Get the file length */
  total = lseek (mp3, 0, SEEK_END);
  if (total < 0)
    {
      perror (fn);
      return -1;
    }
  
  /* Skip id3v1 at the end of the file */
  if (lseek (mp3, -128, SEEK_END) >= 0)
    {
      read (mp3, buf, 3);
      if (!strncmp (buf, "TAG", 3))
	{
	  total -= 128;
	}
    }
  
  /* Get back to the start of the file */
  if (lseek (mp3, 0, SEEK_SET) < 0)
    {
      perror (fn);
      return -1;
    }
  
  return total;
}

/* ======================================================================
   scan_child: Code executed by the child from the fork further down
   ====================================================================== */

void scan_child (int mp3pipe, int respipe)
{
  char *argv[] = { "md5sum", NULL };
  
  /* Make the file descriptors mp3pipe and respipe into stdin and stdout
     and make sure everything went well */
  if (dup2 (mp3pipe, 0) < 0 || dup2 (respipe, 1) < 0)
    {
      perror ("dup2 in child");
      exit (EXIT_FAILURE);
    }
  
  /* Run md5sum */
  execvp ("md5sum", argv);
  
  /* If we ever get here, execvp failed */
  fprintf (stderr, "\n");
  perror ("execvp");
  exit (EXIT_FAILURE);
}

/* ======================================================================
   scan_parent: Code executed by the parent to communicate with the child
   ====================================================================== */

int scan_parent (char *fn, int mp3pipe)
{
  char mp3buf[4096];
  int mp3;
  size_t count = 1;
  off_t skip, total;
  
  /* Open mp3 */
  mp3 = open (fn, O_RDONLY);
  if (mp3 < 0)
    {
      fprintf (stderr, "%-78s\r", "");
      perror (fn);
      close (mp3pipe);
      return -1;
    }
  
  /* Dummy read to check if it's a directory */
  if (read (mp3, mp3buf, 0) < 0)
    {
      fprintf (stderr, "%-78s\r", "");
      perror (fn);
      close (mp3pipe);
      close (mp3);
      return -1;
    }
  
  /* Trim the file from id3-tags and stuff */
  total = trim_file (mp3, fn);
  if (total < 0)
    {
      close (mp3pipe);
      close (mp3);
      return -1;
    }
  
  /* Skip the first SKIP bytes mp3-data, if possible */
  skip = SKIP;
  if (skip + CHECKMIN < total)
    {
      skip = total - CHECKMIN;
      if (skip < 0) skip = 0;
    }
  if (skip && (lseek (mp3, skip, SEEK_CUR) < 0))
    {
      fprintf (stderr, "%-78s\r", "");
      perror (fn);
      close (mp3pipe);
      close (mp3);
      return -1;
    }
  total -= skip;
  
  /* Set the amount of bytes to be read */
  if (CHECKLEN > -1 && total > CHECKLEN)
    {
      total = CHECKLEN;
    }
  
  while (total > 0 && count > 0)
    {
      count = read (mp3, mp3buf, total > 4096 ? 4096 : total);
      if (count > 0)
	{
	  write (mp3pipe, mp3buf, count);
	  total -= count;
	}
    }
  
  /* count indicates an error */
  if (count < 0)
    {
      fprintf (stderr, "%-78s\r", "");
      perror (fn);
      close (mp3pipe);
      close (mp3);
      return -1;
    }
  
  /* This means something fscked up */
  if (total > 0)
    {
      fprintf (stderr, "%-78s\r", "");
      fprintf (stderr, "Warning: file %s is truncated.\n", fn);
    }
  
  close (mp3pipe);
  close (mp3);
  return 0;
}

/* ======================================================================
   scan_file: This takes a filename and creates an md5-sum for it
   ====================================================================== */

int scan_file (char *fn, char *md5)
{
  pid_t child;
  int sstatus, wstatus;
  int respipe[2];
  int mp3pipe[2];
  
  /* Create pipes and make sure they were created */
  if (-1 == pipe (respipe))
    {
      perror ("respipe");
      exit (EXIT_FAILURE);
    }
  if (-1 == pipe (mp3pipe))
    {
      perror ("mp3pipe");
      exit (EXIT_FAILURE);
    }
  
  /* Fork to be able to run md5sum */
  child = fork ();
  switch (child)
    {
    case -1: /* Error while forking */
      perror ("fork");
      exit (EXIT_FAILURE);
      
    case 0: /* This is the child process which should exec md5sum */
      close (mp3pipe[1]);
      close (respipe[0]);
      scan_child (mp3pipe[0], respipe[1]);
      break;
      
    default: /* This is the parent process which should read the mp3 */
      close (mp3pipe[0]);
      close (respipe[1]);
      sstatus = scan_parent (fn, mp3pipe[1]);

      /* Something broke, thus the md5sum is pretty unimportant */
      if (!sstatus)
	{
	  read (respipe[0], md5, MD5LEN);
	  strtok (md5, " ");
	}
      
      /* Close and clean up */
      close (respipe[0]);
      waitpid (child, &wstatus, 0);
      if (!WIFEXITED (wstatus))
	{
	  fprintf (stderr, "Error: Like most parents I have a problem with my "
		   "child.\n Fortunately, I can just quit...\n");
	  exit (EXIT_FAILURE);
	}
    }
  
  return sstatus;
}

/* ======================================================================
   scan_tree: Scans all the files in the mp3-directory
   ====================================================================== */

int scan_tree ()
{
  char fn[FNLEN];
  char md5[MD5LEN];
  int cntr = 0;
  FILE *files;
  FILE *mp3md5;
  
  /* Do a find on the dir containing the files to be processed */
  files = popen ("find " MP3DIR " -xtype f", "r");
  if (!files)
    {
      perror ("scan_tree");
      return EXIT_FAILURE;
    }
  
  /* Create the md5-list */
  mp3md5 = fopen (MP3MD5, "w");
  if (!mp3md5)
    {
      perror (MP3MD5);
      return EXIT_FAILURE;
    }
  
  /* Process the files one by one */
  while (fgets (fn, FNLEN, files))
    {
      /* Filename is not ended by a return, thus it's truncated */
      if (fn[strlen (fn) - 1] != '\n')
	{
	  fprintf (stderr, "Error: Filenamebuffer is %d characters, but "
		   "still too short.\n", FNLEN);
	  return EXIT_FAILURE;
	}
      
      /* Get rid of the newline */
      fn[strlen (fn) - 1] = '\0';
      cntr++;
      fprintf (stderr, "Processing%5d: %-60.60s\r", cntr,
	       strlen (fn) > 60 ? fn + strlen (fn) - 60 : fn);
      
      /* Run the scan and store the result (if successful) */
      if (!scan_file (fn, md5))
	{
	  fprintf (mp3md5, "%s  %s\n", md5, fn);
	}
    }
  
  /* Clean up and stuff */
  fprintf (stderr, "%-78s\r", "");
  fprintf (stderr, "Done: %d %s processed.\n", cntr,
	   cntr == 1 ? "file" : "files");
  pclose (files);
  fclose (mp3md5);
  
  return EXIT_SUCCESS;
}

/* ======================================================================
   compare_files: Compares two files
   ====================================================================== */

int compare_files (char *fn1, char *fn2)
{
  int mp31, mp32;
  off_t total, total1, total2;
  size_t count, count1, count2;
  char mp3buf1[4096], mp3buf2[4096];
  
  /* Open and trim files */
  mp31 = open (fn1, O_RDONLY);
  total1 = trim_file (mp31, fn1);
  mp32 = open (fn2, O_RDONLY);
  total2 = trim_file (mp32, fn2);
  total = 0;
  
  /* Compare files */
  while (total1 > 0 && total2 > 0)
    {
      /* Read files and store data in buffers */
      count1 = total1 > 4096 ? 4096 : total1;
      if (read (mp31, mp3buf1, count1) < count1)
	{
	  fprintf (stderr, "%-78s\r", "");
	  fprintf (stderr, "Error: Read error in %s.\n", fn1);
	  return 0;
	}
      count2 = total2 > 4096 ? 4096 : total2;
      if (read (mp32, mp3buf2, count2) < count2)
	{
	  fprintf (stderr, "%-78s\r", "");
	  fprintf (stderr, "Error: Read error in %s.\n", fn2);
	  return 0;
	}
      
      /* Set count to the lowest amount of read bytes */
      count = count1;
      if (count2 < count) count = count2;
      
      /* If the files aren't equal, return a failure */
      if (memcmp (mp3buf1, mp3buf2, count))
	{
	  return 0;
	}
      
      /* Increase total */
      total += count;
      total1 -= count1;
      total2 -= count2;
    }
  
  /* Check if the files had different length */
  if (total1 < total2)
    {
      fprintf (stderr, "%-78s\r", "");
      fprintf (stderr, "Difference: %s < %s.\n", fn1, fn2);
      return 0;
    }
  if (total2 < total1)
    {
      fprintf (stderr, "%-78s\r", "");
      fprintf (stderr, "Difference: %s > %s.\n", fn1, fn2);
      return 0;
    }
  
  /* Print out a match, if not deselected */
  if (show_matches)
    {
      fprintf (stderr, "%-78s\r", "");
      fprintf (stderr, "Match: %s = %s.\n", fn1, fn2);
    }
  
  return 1;
}

/* ======================================================================
   check_files: Check files against the mp3md5-list
   ====================================================================== */

int check_files (int count, char **fns)
{
  char md5[MD5LEN];
  char md5fn[MD5LEN + FNLEN];
  char *fn;
  FILE *mp3md5;
  int cntr = 0, status;
  
  /* Open the mp3md5-list */
  mp3md5 = fopen (MP3MD5, "r");
  if (!mp3md5)
    {
      perror (MP3MD5);
      return EXIT_FAILURE;
    }
  
  /* If count is zero, use stdin (we need to allocate a buffer then) */
  if (!count) fn = malloc (FNLEN);
  
  /* Get a filename, check it, repeat until done */
  while (cntr < count || (!count && fgets (fn, FNLEN, stdin)))
    {
      status = 0;
      
      /* If count is non zero fns holds the list, otherwise the
	 while-statement above read it from stdin */
      if (count)
	{
	  fn = fns[cntr];
	}
      else
	{
	  if (fn[strlen (fn) - 1] != '\n')
	    {
	      fprintf (stderr, "Error: Filenamebuffer is %d characters, but "
		       "still too short.\n", FNLEN);
	      return EXIT_FAILURE;
	    }
	  fn[strlen (fn) - 1] = '\0';
	}
      
      /* Print progress and md5sum the current file */
      fprintf (stderr, "Processing%5d: %-60.60s\r", cntr + 1,
	       strlen (fn) > 60 ? fn + strlen (fn) - 60 : fn);
      scan_file (fn, md5);
      
      /* Compare the result to the mp3md5-list */
      fseek (mp3md5, 0, SEEK_SET);
      while (fgets (md5fn, MD5LEN + FNLEN, mp3md5) && !status)
	{
	  if (!strncmp (md5, md5fn, strlen (md5)))
	    {
	      /* We have a successful md5-comparement. Now compare the
		 actual files */
	      md5fn[strlen (md5fn) - 1] = '\0';
	      status = compare_files (fn, md5fn + strlen (md5) + 2);
	    }
	}
      
      /* Print out the failed match, if not deselected */
      if (!status && show_no_matches)
	{
	  fprintf (stderr, "%-78s\r", "");
	  fprintf (stderr, "No match: %s.\n", fn);
	}
      
      cntr++;
    }
  
  /* Clean up */
  fprintf (stderr, "%-78s\r", "");
  fprintf (stderr, "Done: %d %s processed.\n", cntr,
	   cntr == 1 ? "file" : "files");
  
  return EXIT_SUCCESS;
}

int main (int argc, char **argv)
{
  extern int optind;
  
  /* Check options and take actions */
  while (1)
    {
      int c = getopt (argc, argv, "hmns");
      
      if (c == -1) break;
      
      switch (c)
	{
	case 'h':
	  fprintf (stderr, "mp3md5sum usage:\n"
		   "  -h  This help\n"
		   "  -m  Don't show matches\n"
		   "  -n  Don't show no matches\n"
		   "  -s  Scan the dir " MP3DIR
		   " and store all md5sums in the file " MP3MD5 "\n");
	  return EXIT_SUCCESS;
	  
	  /* Handle flags */
	case 'm':
	  show_matches = 0;
	  break;
	case 'n':
	  show_no_matches = 0;
	  break;
	  	
	  /* Generate the mp3md5-list */
	case 's':
	  return scan_tree ();
	  
	  /* Handle errors */
	case '?':
	  fprintf (stderr, "Houston, we have a problem.\n");
	  return EXIT_FAILURE;
	  
	default:
	  fprintf (stderr, "Character %c not recognized.\n", c);
	  return EXIT_FAILURE;
	}
    }
  
  /* Check files to the mp3md5-list */
  return check_files (argc - optind, argv + optind);
}
